// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of omnitest.data;

/// Expected events, requests, responses and HTTP requests
class OmnitestExpect {
  /// Expected events
  List<OmnitestEvent> expectedEvents = new List<OmnitestEvent>();

  /// Expected requests
  List<OmnitestExpectedRequest> expectedRequests =
      new List<OmnitestExpectedRequest>();

  /// Expected (final) responses
  List<OmnitestResponse> expectedResponses = new List<OmnitestResponse>();

  /// Expected HTTP requests
  List<OmnitestHttpRequest> expectedHttpRequests =
      new List<OmnitestHttpRequest>();

  /// Load data from a YAML map.
  OmnitestExpect.fromYaml(
      YamlList yamlList, Map<String, OmnibuildMessage> messageSpec) {
    // Iterate through top-level messages.
    yamlList.forEach((dynamic item) {
      if (item is YamlMap) {
        YamlMap yamlMap = item;
        String omnibusId = getOmnibusId(messageSpec.keys, yamlMap);
        if (omnibusId != null) {
          switch (messageSpec[omnibusId].type) {
            case OmnibusMessageType.event:
              // Parse message as event and store it in $expectedEvents.
              expectedEvents.add(new OmnitestEvent(
                  omnibusId, resolveMessageData(yamlMap, omnibusId)));
              break;

            case OmnibusMessageType.response:
              // Parse message as response and store it in $expectedResponses.
              expectedResponses.add(new OmnitestResponse(
                  omnibusId, resolveMessageData(yamlMap, omnibusId)));
              break;

            case OmnibusMessageType.request:
              // Requests also have expected responses, therefore they are
              // stored using the OmnitestRequest class.
              var request = new OmnitestExpectedRequest.fromYaml(
                  omnibusId, yamlMap, messageSpec);

              // Add expectedHttpRequests from this request to our
              // expectedHttpRequests.
              expectedHttpRequests.addAll(request.expectedHttpRequests);
              expectedRequests.add(request);
              break;
          }
        } else {
          throw new OmnitestException.noOmnibusIdKey(yamlMap);
        }
      } else if (item is String) {
        expectedHttpRequests.add(new OmnitestHttpRequest.fromString(item));
      }
    });

    // Check for duplicate requests
    //
    // Duplicate events and responses are allowed. However, duplicate requests
    // would either point to stateful requests which are highly discouraged,
    // or a useless repeated request.
    Set<OmnibusRequest> reqs = new Set<OmnibusRequest>();
    for (var req in expectedRequests) {
      if (!reqs.contains(req.data)) {
        reqs.add(req.data);
      } else {
        // This request is already contained in $reqs; throw an error.
        throw new OmnitestException.yaml(
            'You cannot put duplicate requests in expect.');
      }
    }
  }

  /// Find the first [OmnitestExpectedRequest] for the given [OmnitestRequest].
  OmnitestExpectedRequest findRequest(OmnitestRequest request) {
    return expectedRequests.firstWhere(
        (OmnitestExpectedRequest r) => request == r.data,
        orElse: () => null);
  }

  /// Get raw array of expected requests.
  List<OmnitestRequest> getExpectedRequests() {
    return new List<OmnitestRequest>.generate(
        expectedRequests.length, (int i) => expectedRequests[i].data);
  }
}
