// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of omnitest.data;

/// Single test file data structure
class OmnitestFile {
  /// Description YAML key
  static const descriptionKey = 'description';

  /// Message spec YAML key
  static const messageSpecKey = 'message-spec';

  /// Expect onload YAML key
  static const expectOnloadKey = 'expect-onload';

  /// Use cases YAML key
  static const useCasesKey = 'use-cases';

  /// General description
  String description;

  /// Message details per omnibusId (parsed from the message-spec list)
  Map<String, OmnibuildMessage> messageSpec =
      new Map<String, OmnibuildMessage>();

  /// Expected activity when the module is loaded.
  OmnitestExpect expectOnload;

  /// List of use cases that are defined in this file
  List<OmnitestUseCase> useCases = new List<OmnitestUseCase>();

  /// Load data from a YAML map.
  Future parseYaml(String yamlString) async {
    var yamlData = loadYaml(yamlString);

    // Read description.
    description =
        yamlData.containsKey(descriptionKey) ? yamlData[descriptionKey] : '';

    // Load own package name.
    String thisPackage =
        loadYaml(new File('./pubspec.yaml').readAsStringSync())['name'];

    // Parse message spec.
    var pathMatcher = new RegExp(r'.v(\d+).yaml$');
    var packageMatcher = new RegExp(r'package:([^/]+)\/');
    if (yamlData.containsKey(messageSpecKey) &&
        yamlData[messageSpecKey] is YamlList) {
      for (var uri in yamlData[messageSpecKey]) {
        var r = pathMatcher.firstMatch(uri);

        // Only proceed if the file URI meets the naming requirements.
        // (name.v#.yaml)
        if (r != null) {
          // Resolve package name.
          var packageName = thisPackage;
          if (uri.startsWith('package:')) {
            var pr = packageMatcher.firstMatch(uri);
            packageName = pr != null ? pr.group(1) : packageName;
          }

          // Parse resource.
          var resource = new Resource(uri);
          var yamlString = await resource.readAsString();
          var spec = new OmnibuildSpecFile.fromYaml(
              yamlString, int.parse(r.group(1)), packageName);

          // Insert all messages into messageSpec.
          spec.messages.forEach((OmnibuildMessage message) {
            messageSpec.putIfAbsent(message.omnibusId, () => message);
          });
        } else {
          // Throw error.
          throw new OmnitestException.yaml(
              '$uri does not meet the naming convention (name.v#.yaml).');
        }
      }
    }

    // Check if the message spec contains any reserved names.
    if (reservedKeys
        .toSet()
        .intersection(messageSpec.keys.toSet())
        .isNotEmpty) {
      throw new OmnitestException.yaml(
          '$reservedKeys can not be used as omnibusId.');
    }

    // Parse expect onload.
    expectOnload = new OmnitestExpect.fromYaml(
        yamlData.containsKey(expectOnloadKey)
            ? yamlData[expectOnloadKey]
            : new YamlList(),
        messageSpec);

    // Loop through use cases.
    if (yamlData.containsKey(useCasesKey) &&
        yamlData[useCasesKey] is YamlList) {
      yamlData[useCasesKey].forEach((YamlMap useCase) {
        useCases.add(new OmnitestUseCase.fromYaml(useCase, messageSpec));
      });
    }
  }
}
