// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of omnitest.exceptions;

/// JSON encoder for writing fail messages.
final JsonEncoder _jsonEncoder = new JsonEncoder.withIndent('  ');

/// AnsiPen to write red text.
final redPen = new AnsiPen()..red();

/// Split [a] into lines and prefix them with [b]. Returns all lines joined
/// together using line endings.
String _prefixEachLine(String str, String prefix) {
  var out = new StringBuffer();
  var lines = LineSplitter.split(str);
  lines.forEach((String line) {
    out.writeln('$prefix$line');
  });
  return out.toString();
}

/// Error class used for internal error throwing.
class OmnitestException extends Error {
  /// Error message
  String message;

  /// Default constructor
  OmnitestException(this.message);

  /// Render YAML error
  OmnitestException.yaml(String error) {
    message = redPen('There was an error in your YAML test file:\n$error\n');
  }

  /// Render no omnibusId in message map found error
  OmnitestException.noOmnibusIdKey(YamlMap data) {
    message = redPen('''
There was an error in your YAML test file:
No valid omnibusId was found in the following message:
${_jsonEncoder.convert(data)}
''');
  }

  /// Render unexpected events error.
  OmnitestException.unexpectedEvents(
      List<OmnibusEvent> expected, List<OmnibusEvent> actual) {
    // Render message
    message = redPen(_msgUnexpectedMessages.renderString({
      'something': 'events',
      'expected': _generateMustacheList(expected),
      'actual': _generateMustacheList(actual)
    }));
  }

  /// Render unexpected requests error.
  OmnitestException.unexpectedRequests(
      List<OmnibusRequest> expected, List<OmnibusRequest> actual) {
    // Render message
    message = redPen(_msgUnexpectedMessages.renderString({
      'something': 'requests',
      'expected': _generateMustacheList(expected),
      'actual': _generateMustacheList(actual)
    }));
  }

  /// Render unexpected HTTP requests error.
  OmnitestException.unexpectedHttpRequests(List expected, List actual) {
    // Render message
    message = redPen(_msgUnexpectedHttpRequests
        .renderString({'expected': expected, 'actual': actual}));
  }

  /// Render unexpected responses error.
  OmnitestException.unexpectedResponses(OmnibusRequest request,
      List<OmnibusResponse> expected, List<OmnibusResponse> actual) {
    // Render message
    message = redPen(_msgUnexpectedResponses.renderString({
      'request': {
        'omnibusId': request.omnibusId,
        'jsonData':
            _prefixEachLine(_jsonEncoder.convert(request.toPds()), '    ')
                .substring(4) // Remove indent on first line
      },
      'expectedResponses': _generateMustacheList(expected),
      'actualResponses': _generateMustacheList(actual)
    }));
  }

  /// Render unexpected request error.
  OmnitestException.unexpectedRequest(OmnibusRequest request) {
    message = redPen(_msgUnexpectedRequest.renderString({
      'omnibusId': request.omnibusId,
      'jsonData': _prefixEachLine(_jsonEncoder.convert(request.toPds()), '    ')
          .substring(4) // Remove indent on first line
    }));
  }

  /// toString should return the message.
  String toString() => message;

  /// Generate mustache list.
  List _generateMustacheList(List<OmnibusMessage> list) {
    var newList = new List<Map<String, String>>();
    list.forEach((OmnibusMessage item) {
      newList.add({
        'omnibusId': item.omnibusId,
        'jsonData': _prefixEachLine(_jsonEncoder.convert(item.toPds()), '    ')
            .substring(4) // Remove indent on first line
      });
    });
    return newList;
  }
}

/// Template for _OmniError.unexpectedEvents and _OmniError.unexpectedRequests
final _msgUnexpectedMessages = new Template('''
The published {{ something }} were not as expected

> Expected {{ something }}:
{{# expected }}
- omnibusId: {{ omnibusId }}
  data: {{{ jsonData }}}
{{/ expected }}
> Actual {{ something }}:
{{# actual }}
- omnibusId: {{ omnibusId }}
  data: {{{ jsonData }}}
{{/ actual }}
''');

/// Template for _OmniError.unexpectedHttpRequests
final _msgUnexpectedHttpRequests = new Template('''
The fired HTTP requests were not as expected

> Expected HTTP requests:
{{# expected }}
- method: {{ method }}
  url: {{{ url }}}
{{/ expected }}
> Actual HTTP requests:
{{# actual }}
- method: {{ method }}
  url: {{{ url }}}
{{/ actual }}
''');

/// Template for _OmniError.unexpectedResponses
final _msgUnexpectedResponses = new Template('''
The module published unexpected responses

> Request:
  omnibusId: {{ request.omnibusId }}
  data: {{{ request.jsonData }}}
> Expected responses from module:
{{# expectedResponses }}
- omnibusId: {{ omnibusId }}
  data: {{{ jsonData }}}
{{/ expectedResponses }}
> Actual responses from module:
{{# actualResponses }}
- omnibusId: {{ omnibusId }}
  data: {{{ jsonData }}}
{{/ actualResponses }}
''');

/// Template for _OmniError.unexpectedRequest
final _msgUnexpectedRequest = new Template('''
A request was published but not expected

omnibusId: {{ omnibusId }}
data: {{{ jsonData }}}
''');
