// Copyright (c) 2016, Herman Bergwerf. All rights reserved.
// Use of this source code is governed by an AGPL-3.0-style license
// that can be found in the LICENSE file.

part of omnitest.test;

/// Run all tests described in [tests].
Future runTests(OmnitestFile tests, ModuleMain module) async {
  // Loop through use cases.
  for (var i = 0; i < tests.useCases.length; i++) {
    await runZoned(() async {
      // Used to check if a given request is expected.
      OmnitestExpect currentExpect;

      // Important note on stateful requests
      // -----------------------------------
      // The usage of stateful requests cannot be tested with omnitest since
      // you can only define a specific request once per message in the
      // message chain, and not multiple times with different responses. This
      // is by design, your code should never have to rely on state data in
      // the modules behind the messages.

      // Create new test fabric.
      var fabric = new OmnitestFabric(
          (OmnitestRequest request, List<OmnitestResponse> moduleResponses) {
        OmnitestExpectedRequest requestInformation =
            currentExpect.findRequest(request);
        if (requestInformation != null) {
          // Check if the module responses are as expected.
          if (!orderlessListCompare(
              moduleResponses, requestInformation.moduleResponses)) {
            throw new OmnitestException.unexpectedResponses(
                request, requestInformation.moduleResponses, moduleResponses);
          }

          // Return the mocked responses.
          return requestInformation.mockedResponses;
        } else {
          throw new OmnitestException.unexpectedRequest(request);
          return [];
        }
      });

      // First, capture the onload activity.
      currentExpect = tests.expectOnload;
      fabric.clearLogs();

      // Await module onload.
      await Future.wait([
        module(fabric.bus, fabric.httpLogger),
        module(fabric.secondBus, new http.Client())
      ]);

      /// Validate onload activity.
      fabric.check(tests.expectOnload);

      // Run message chain.
      for (var message in tests.useCases[i].messageChain) {
        currentExpect = message.expect;
        await testMessage(message, fabric);
      }
    }, onError: (e, stackTrace) {
      var nr = i + 1;
      fail('''
In testcase #$nr:
Description: ${tests.useCases[i].description}

$e
${e is OmnitestException ? '' : '\n$stackTrace\n'}''');
    });
  }
}

/// Evaluate the outcome of a single message on the given fabric.
Future testMessage(OmnitestChainMessage message, OmnitestFabric fabric) async {
  fabric.clearLogs();

  // Publish message.
  if (message.data.type == OmnibusMessageType.event) {
    // Await all event handlers.
    await (await fabric.bus.publishEvent(message.data)).onComplete();
  } else {
    // Block the processing of the tested request by the fabric.
    fabric.requestHandler.skipOnce();

    // Await all responses.
    var responses =
        await (await fabric.bus.publishRequest(message.data)).onComplete();

    // Convert all responses to OmnitestResponse to make them comparable.
    var omnitestResponses = new List<OmnitestResponse>();
    responses.forEach((OmnibusResponse response) => omnitestResponses
        .add(new OmnitestResponse(response.omnibusId, response.toPds())));

    // Check responses.
    if (!orderlessListCompare(
        message.expect.expectedResponses, omnitestResponses)) {
      throw new OmnitestException.unexpectedResponses(
          message.data, message.expect.expectedResponses, responses);
    }
  }

  // Validate module activity.
  fabric.check(message.expect);
}
